
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Functions to look for changes in the correlation matrix of a multivariate time series
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

######################################################################################################
######################################################################################################
# The main algorithm of the procedure. It provides with the results for binary segmentation and binary
# segmentation with refinement step
######################################################################################################
######################################################################################################

######################################################################################################
# The function BSWRS.CorMat is the main function of the procedure for detecting multiple changepoints
# in the correlation matrix of a multivariate time series using binary segmentation and binary 
# segmentation with refinement step.
# The inputs are:
# 1. X, the matrix of size n x p with the time series (n, sample size, and p, number of series)
# 2. B, the number of bootstrap samples
# 3. gen.val, are the 10000 values generated from the distribution of the supremum of the L1 norm of the
# vector of brownian bridges
# The outputs are:
# 1. BS.ch.po, the location of the changepoints with binary segmentation
# 2. BS.st.ch.po, the associated statistics with binary segmentation
# 3. BS.ch.po.st, the estimated changepoints (in the interval (0,1]) with binary segmentation
# 4. BSWRS.ch.po, the location of the changepoints with binary segmentation with refinement step
# 5. BSWRS.st.ch.po, the associated statistics with binary segmentation with refinement step
# 6. BSWRS.ch.po.st, the estimated changepoints (in the interval (0,1]) with binary segmentation with 
# refinement step
######################################################################################################

BSWRS.CorMat <- function(X,B,gen.val){
  
  n <- dim(X)[1] # Sample size
  p <- dim(X)[2] # Dimension
  a.0 <- .05 # Initial level

  # The first part of the algorithm is binary segmentation.
  
  ch.po <- c() # Set detected changepoints
  au.ch.po <- c(0,ch.po,n) # Augmented changepoints
  nu.ch.po <- 0 # Set number of detected changepoints
  st.ch.po <- c() # Set associated statistics
  in.in <- 1 # Set indicator to look changepoints in each subinterval
  
  iter <- 0 # Set iteration
  look <- 1 # Set indicator to look
  
  max.st <- c() # Set maxima of test in each subinterval
  ch.max.st <- c() # Set locations of maxima of test in each subinterval
  
  while ((look == 1) && (iter < 10)){ # Control a large number of iterations
    iter <- iter + 1 # New iteration 
    for (i in 1 : (nu.ch.po + 1)){ # Repeat for every subinterval
      if (in.in[i]==1){
        n.ini <- au.ch.po[i] + 1 # First observation
        n.end <- au.ch.po[i+1] # Last observation
        st <- rep(0,n) # Set values of statistics
        st.NoE <- rep(0,n)
        aux <- test.bootstrap(X[n.ini:n.end,],B)
        st[n.ini:n.end] <- aux$teststatistic # Computation of the statistic in the subinterval
        st.NoE[n.ini:n.end] <- aux$teststatistic.NoE # Computation of the statistic in the subinterval
        max.st[i] <- max(st) # Maximum stat in the subinterval
        ch.max.st[i] <- which.max(st.NoE) # Location of the previous maximum
      }
    }  
    
    max.max.st <- max(max.st) # Maximum of the statistics in this iteration in all the subintervals
    wh.max.max.st <- which.max(max.st)
    ch.po.max.st <- ch.max.st[wh.max.max.st] # Estimation of the new changepoint
    
    p.0 <- (1 - a.0)^(1 / (length(ch.po) + 1)) # Probability according with the number of candidates
    cv <- quantiles.finder(prob=p.0,gen.val) # Find the corresponding quantile with the gen.val vector
    
    if ((max.max.st > cv) && (ch.po.max.st>5)){ # If the statistic is significant, then define the new candidate
      so.ch.po <- sort(c(ch.po,ch.po.max.st),index.return=T) # New vector of changepoints
      ch.po <- so.ch.po$x
      au.ch.po <- c(0,ch.po,n) # Augmented changepoints
      nu.ch.po <- length(ch.po) # Set number of detected changepoints
      st.ch.po <- c(st.ch.po,max.max.st)[so.ch.po$ix] # Set associated statistics
      if (min(diff(au.ch.po)) < 5){
        look <- 0 # Stop the search of new changepoints if a subinterval is very small
      }else{
        in.in <- rep(0,nu.ch.po+1)
        in.in[wh.max.max.st:(wh.max.max.st+1)] <- 1
        
        max.st.aux <- rep(0,nu.ch.po+1)
        max.st.aux[-(wh.max.max.st:(wh.max.max.st+1))] <- max.st[-wh.max.max.st]
        max.st <- max.st.aux
        
        ch.max.st.aux <- rep(0,nu.ch.po+1)
        ch.max.st.aux[-(wh.max.max.st:(wh.max.max.st+1))] <- ch.max.st[-wh.max.max.st]
        ch.max.st <- ch.max.st.aux
      }
    }else{
      look <- 0 # Stop the search of new changepoints
    }  
  }  
  ch.po.st <- ch.po / n
  
  BS.ch.po <- ch.po
  BS.st.ch.po <- st.ch.po
  BS.ch.po.st <- ch.po.st

  # The second part of the algorithm is refinement step.
  
  p.0 <- 1 - a.0 # Fixed probability
  cv <- quantiles.finder(prob=p.0,gen.val) # Find the corresponding quantile with the gen.val vector
  
  ch.po.aux <- ch.po
  st.ch.po.aux <- st.ch.po
  ch.po.st.aux <- ch.po.st
  
  look <- 1
  iter <- 0
  while ((look == 1) && (iter < 10)){
    iter <- iter + 1
    l.ch.po.aux <- length(ch.po.aux)
    if (l.ch.po.aux == 0){
      BSWRS.ch.po <- c()
      BSWRS.st.ch.po <- c()
      BSWRS.ch.po.st <- c()
      look <- 0
    }
    if (l.ch.po.aux == 1){
      BSWRS.ch.po <- ch.po.aux
      BSWRS.st.ch.po <- st.ch.po.aux
      BSWRS.ch.po.st <- ch.po.st.aux
      look <- 0
    }
    if (l.ch.po.aux > 1){
      au.ch.po.aux <- c(0,ch.po.aux,n)
      ch.po.new <- c()
      st.ch.po.new <- c()
      for (i in 1 : l.ch.po.aux){ # Repeat for every subinterval
        n.ini <- au.ch.po.aux[i] + 1 # First observation
        n.end <- au.ch.po.aux[i+2] # Last observation
        st <- rep(0,n) # Set values of statistics
        st.NoE <- rep(0,n)
        aux <- test.bootstrap(X[n.ini:n.end,],B) # Computation of the statistic in the subinterval
        st[n.ini:n.end] <- aux$teststatistic # Computation of the statistic in the subinterval
        st.NoE[n.ini:n.end] <- aux$teststatistic.NoE # Computation of the statistic in the subinterval
        max.st <- max(st) # Maximum stat in the subinterval
        ch.max.st <- which.max(st.NoE) # Location of the previous maximum
        if (max.st > cv){
          ch.po.new <- c(ch.po.new,ch.max.st)
          st.ch.po.new <- c(st.ch.po.new,max.st)
        }
      }
      l.ch.po.new <- length(ch.po.new)
      if (l.ch.po.new == l.ch.po.aux){
        if (sum(ch.po.new==ch.po.aux)==l.ch.po.new){
          BSWRS.ch.po <- ch.po.new
          BSWRS.st.ch.po <- st.ch.po.new
          BSWRS.ch.po.st <- ch.po.new / n
          look <- 0
        }else{
          ch.po.aux <- ch.po.new
          st.ch.po.aux <- st.ch.po.new
          ch.po.st.aux <- ch.po.aux / n
        }
      }else{
        ch.po.aux <- ch.po.new
        st.ch.po.aux <- st.ch.po.new
        ch.po.st.aux <- ch.po.aux / n
      }
    }
  }  
  
  if (iter == 10){
    BSWRS.ch.po <- ch.po.new
    BSWRS.st.ch.po <- st.ch.po.new
    BSWRS.ch.po.st <- ch.po.new / n
  }
  
  return(list("BS.ch.po"=BS.ch.po,"BS.st.ch.po"=BS.st.ch.po,"BS.ch.po.st"=BS.ch.po.st,
              "BSWRS.ch.po"=BSWRS.ch.po,"BSWRS.st.ch.po"=BSWRS.st.ch.po,"BSWRS.ch.po.st"=BSWRS.ch.po.st))
  
}

######################################################################################################
######################################################################################################
# Functions for the test based on the bootstrap approximation of E
######################################################################################################
######################################################################################################

######################################################################################################
# The function test.bootstrap computes the value of the statistic using the bootstrap procedure. 
# The inputs are:
# 1. X, a matrix of size n x p where n is the sample size and p is the dimension of the time series
# 2. B, the number of bootstrap replications
# The outputs are:
# 1. teststatistics, a vector with the computed statistics for each time point in X
# 2. teststatistics.NoE, a vector with the computed statistics without the bootstrap estimate of the 
# matrix E, for each time point in X
######################################################################################################

test.bootstrap <- function(X,B){
  
  n <- dim(X)[1] # Sample size
  p <- dim(X)[2] # Number of time series
  
  l <- floor(n^(1/4)) # Length of the blocks

  R <- cor(X) # Correlation matrix of the whole dataset
  Dc <- matrix(0,nrow=n,ncol=p*(p-1)/2) # Initialize the value of the statistics
  for (k in 2 : n){
      DR <- cor(X[1:k,]) - R # Correlation matrix until time k minus correlation matrix of the whole dataset
      Dc[k,] <- k / sqrt(n) * DR[lower.tri(DR)] # Vectors needed to compute the statistics
  }
  
  Bl <- c()
  for (i in 1:(n-l+1)){Bl[[i]] <- X[i:(i+l-1),]} # List with blocks of data
  
  Eh <- NA
  while (any(is.na(Eh))){
    cor.boot <- matrix(0,nrow=B,ncol=p*(p-1)/2) # Initialize 
    for (b in 1 : B){ # Initialize the bootstrap
      aux <- NA
      while (any(is.na(aux))){
        ind <- sort(sample(1:(n-l+1),n/l,replace=T)) # Needed to select the blocks
        XBl <- do.call("rbind",Bl[ind]) # Merge by rows the selected blocks
        aux <- apply(XBl,2,sd)
        if (any(is.na(aux)) || any(aux==0)){
          RXBl <- c()
        }else{
          RXBl <- cor(XBl) # Compute correlation matrix of the bootstrapped series 
          cor.boot[b,] <- sqrt(n) * RXBl[lower.tri(RXBl)]
        }
      }
    }  
    Eh <- cov(cor.boot) # Covariance of the rows in cor.boot
  }  
  aux <- eigen(Eh)
  evectors <- aux$vectors
  evalues <- aux$values
  if (any(evalues==0)){evalues <- evalues + 10^(-10)}
  Ehi12 <- evectors %*% diag(1/sqrt(abs(evalues))) %*% t(evectors) # Needed to compute the statistics
  
  teststatistic <- apply(apply(Dc,1,function(x) abs(Ehi12 %*% x)),2,sum) # Computation of the statistics
  teststatistic.NoE <- apply(apply(Dc,1,function(x) abs(x)),2,sum) # Computation of the statistics without normalizing constant
  
  return(list("teststatistic"=teststatistic,"teststatistic.NoE"=teststatistic.NoE))
  
}

######################################################################################################
# The function generated.values.test.bootstrap generates values from the supremum of p*(p-1)/2 Brownian
# Bridges that are needed to compute quantiles.
# The inputs is:
# 1. p, the dimension of the time series
# The output is:
# 1. gen.val, the generated values
######################################################################################################

generated.values.test.bootstrap <- function(p,nsam=10000){
  
  m <- 1000 # Number of points in the grid in the [0,1] interval
  dm <- 1 / (m - 1) # Step length in the grid
  dms <- seq(0,1,length.out=m) # Sequence of points
  gen.val <- matrix(0,nrow=nsam,ncol=1) # Initialize the vector with generated values
  for (k in 1 : nsam){ # Generation of the Brownian motion and the associated Brownian bridge
    BM <- matrix(0,nrow=m,ncol=p*(p-1)/2) # Initialize Brownian motion
    BB <- matrix(0,nrow=m,ncol=p*(p-1)/2) # Initialize Brownian bridge
    for (j in 1:(p*(p-1)/2)){
      BM[1,j] <- 0
      for (i in 2 : m){
        BM[i,j] <- BM[i-1,j] + rnorm(1,mean=0,sd=sqrt(dm))
      }
      BB[,j] <- BM[,j] - dms * BM[m,j]
    }
    gen.val[k] <- max(apply(abs(BB),1,sum)) # Critical values generated
  }
  gen.val <- sort(gen.val) # Sort the critical values
  
  return(gen.val)
  
}

######################################################################################################
# The function quantiles.finder computes quantiles of a set of values for a certain probability
# The inputs is:
# 1. prob, the value at which we want to compute the quantile
# 2. gen.val, the values generated previously from the distribution
# The output is:
# 1. quant, the required quantile
######################################################################################################

quantiles.finder <- function(prob,gen.val){
  
  lgv <- length(gen.val) # Length of the vector of generated values
  quant <- gen.val[floor(prob * lgv)] # Compute quantile
  
  return(quant)
  
}

######################################################################################################
######################################################################################################
# Functions for simulating models with changes in correlations (or not)
######################################################################################################
######################################################################################################

########################################################################################################
# The function Sim_SBEKK simulates a time series following a scalar BEKK(1,1) model with a changing correlation 
# matrix (or not). The mean and variances are constants.
# The inputs are:
# 1. n, the sample size
# 2. p, the number of time series
# 3. rhos, a list with the correlation matrices
# 4. param, the vector of parameters
# 5. ncs, the vector containing the changepoints
# 6. dis, the distribution (1 for Gaussian, 2 for Student-t with 3 degrees of freedom)
# The outputs are:
# 1. X, the matrix of size nxp with the time series
########################################################################################################

Sim_SBEKK <- function(n,p,rhos,param,ncs,dis){
  
  mu <- param[1:p]
  sds <- param[(p+1):(2*p)]
  al <- param[2*p+1]
  be <- param[2*p+2]
  
  m <- n + 100
  X <- matrix(0,nrow=p,ncol=m)
  
  nch <- length(ncs)
  nc <- c(0,ncs+100,n+100)
    
  if (dis == 1){
    ept <- mvrnorm(n=m,mu=rep(0,p),Sigma=diag(p))
  }else if (dis == 2){
    ept <- rmvt(n=m,sigma=diag(p)*(5-2)/5,df=5)
  }
  
  ept <- t(ept)
  
  Ht <- diag(sds) %*% rhos[[1]] %*% diag(sds)
  Ht.eig <- eigen(Ht)
  Ht.sqrt <- Ht.eig$vectors %*% diag(sqrt(Ht.eig$values)) %*% t(Ht.eig$vectors)
  
  X[,1] <- Ht.sqrt %*% ept[,1] 
    
  for (i in 2 : m){
    
    if (nch==0){
      R <- rhos[[1]]
    }else{
      R <- rhos[[sum(i>nc)]]  
    }  
    H <- diag(sds) %*% R %*% diag(sds)
    
    Ht <- (1 - al^2 - be^2) * H + al^2 * X[,i-1] %*% t(X[,i-1]) + be^2 * Ht
    Ht.eig <- eigen(Ht)
    Ht.sqrt <- Ht.eig$vectors %*% diag(sqrt(Ht.eig$values)) %*% t(Ht.eig$vectors)
    
    X[,i] <- Ht.sqrt %*% ept[,i]
    
  }
  
  X <- t(X[,101:m])
  
  return(X)
  
}
